/***
 * llstat.c -- some routines for histograms, average, random numbers
***/

/*** Implementation of statistical analysis of data   ***/
/*** A.K. Hartmann January 1995                       ***/
/*** if _LLSTAT_SEC_ is defined at compile time, then ***/
/*** additional security checks are included:         ***/
/*** 1. Each stat-Pointer is tested, wether it is     ***/
/***    NULL or does not point to a stat.             ***/
/*** Version 0.4.5.0  014.04.2005                      ***/

/*** Vers.History:                                                    ***/
/*** V 0.2.0.3 #define _INCLUDE_HPUX_SOURCE for cl1-Cluster (CHANGE1) ***/
/*** V 0.2.1.0 comment for cs2html added                              ***/
/*** V 0.2.1.1 check from,to,nbask in llstat_new (CHANGE2)            ***/
/*** V 0.2.1.2 DEBUG-Option added                                     ***/
/*** V 0.2.1.3. #ifdef AUF_PARIX -> #ifdef PARIX                      ***/
/*** V 0.2.2.0 stderror to mean value in llstat_write() added.        ***/
/*** V 0.2.3.0 llstat_max(),llstat_error() added.                     ***/
/*** V 0.2.3.1 new llstat_error()                                     ***/
/*** V 0.3.0.0 correlations added                                     ***/
/*** V 0.3.0.1 return in llcorr_new added                             ***/
/*** V 0.3.1.0 llcorr_add_stat() added                                ***/
/*** V 0.3.2.0 llstat_int() added                                     ***/
/*** V 0.3.3.0 nicer prinout of stats and corrs                       ***/
/*** V 0.3.3.1 error in llcorr_insert() removed (CHANGE3)             ***/
/*** V 0.3.4.0 llstat_binder() added.                                 ***/
/*** V 0.3.4.1 llstat_write() writes "%f %f"-format                   ***/
/*** V 0.3.5.0 llstat_insert_mult() added, higher moments             ***/
/*** V 0.3.6.0 llstat_binder2() added.                                ***/
/*** V 0.3.6.1 output of moments in llstat_write() added.             ***/
/*** V 0.3.7.0 llstat_moment(), llstat_error_moment, llcorr_moment()  ***/
/*** V 0.3.7.1 in llcorr_add_stat(): only add non-empty stats         ***/
/*** V 0.3.8.0 stat->table (int *) -> (double *)                      ***/
/*** V 0.3.8.1 normalization in llstat_write(),llstat_int(),          ***/
/***           llstat_w_pc(), llcorr_add_stat() changed               ***/
/*** V 0.3.9.0 llstat_write_log() added                               ***/
/*** V 0.4.0.0 llcorr_write_log() added (26.7.2000)                   ***/
/*** V 0.4.0.1 calc. of bin pos. for _write_log() changed (29.8.00)   ***/
/*** V 0.4.1.0 errorbar for printing, insert_mult() non-int (9.3.01)  ***/
/*** V 0.4.2.0 jackknife estimator added (23.4.01)                    ***/
/*** V 0.4.3.0 print #samples at llcorr_write() (24.4.01)             ***/
/*** V 0.4.4.0 jackknife for moments and llstat_binder2() (2.5.01)    ***/
/*** V 0.4.4.1 wrong free-order in llstat_delete() removed (13.6.01)  ***/
/*** V 0.4.4.2 llcorr_write_log() corrected (8.10.02)                 ***/
/*** V 0.4.4.3 variance=value^2 if only one value (19.2.03)           ***/
/*** V 0.4.4.4 llstat_write_log works for no stdout (1.7.03)          ***/
/*** V 0.4.5.0 llcorr_mean(),llstat_dens() added (14.4.05)            ***/

#include <stdio.h>
#include <stdlib.h>

#ifdef AUF_CL1     /*CHANGE1*/
#define _INCLUDE_HPUX_SOURCE
#endif

#ifdef PARIX
#include <gnu/time.h>
#else
#include <sys/time.h>
#endif

#include <math.h>

#ifdef M_PI
#define PI M_PI
#else
#define PI 3.1415927
#endif

#ifndef RAND_MAX
#define RAND_MAX 2147483647.0
#endif

#ifdef DEBUG
#include "ldebug.h"
#endif

#include "makros_ah.h"
#include "llstat.h"

/* function commentary header (empty): */

/*******************  *****************/
/**  **/
/** PARAMETERS: (*)= return-paramter         **/
/**  **/
/** RETURNS:                               **/
/**  **/
/******************************************/


/** creates a stat-Element, where the empirical ditribution **/
/** table covers the range ['from', 'to'] and is divied     **/
/** into 'n_bask' baskets.                                  **/
/** No jackknife possible                                   **/
/** RETURNS: pointer to stat-Element or NULL if no memory.  **/
llstat_t *llstat_new(double from, double to, int n_bask)
{
  llstat_t *stat;
  int t;

  TF_MALLOC_AH(stat, llstat_t, "Statistic");
  if(stat != NULL)
  {
	if(to < from) /* (CHANGE2) */
	{
	  double tmp;
	  tmp = to; to = from; from = tmp;
	  fprintf(stderr, "WARNING: exchanging from, to in llstat_new\n");
	}

    stat->from = from;
    stat->to = to;
	if( n_bask <= 0)  /* (CHANGE2) */
	{
	  n_bask = 10;
	  fprintf(stderr, "WARNING: setting n_bask=10 in llstat_new()\n");
	}
    stat->delta = (to-from)/(double) n_bask;
    stat->n_bask = n_bask;
    stat->low = 0;
    stat->high = 0;
    stat->num_jackknife = 0;
    stat->act_jackknife = 0;
    stat->sum_jackknife = NULL;
    stat->magic = _LLSTAT_MAGIC_;
    for(t=0; t< _LLSTAT_NOM_ ; t++)
      stat->sum[t] = 0.0;
    TFN_MALLOC_AH(stat->table, n_bask, double, "Statistic");
    if(stat->table == NULL)
    {
      free(stat);
      stat = NULL;
    }
    else
      for(t=0; t<n_bask; t++)
	stat->table[t] = 0;
  }
  return(stat);
}

/** creates a stat-Element, where the empirical ditribution **/
/** table covers the range ['from', 'to'] and is divied     **/
/** into 'n_bask' baskets.                                  **/
/** With jackknife possible                                 **/
/** RETURNS: pointer to stat-Element or NULL if no memory.  **/
llstat_t *llstat_new_jackknife(double from, double to, int n_bask,
			       int num_jackknife)
{
  llstat_t *stat;
  int t, moment;

  TF_MALLOC_AH(stat, llstat_t, "Statistic");
  if(stat != NULL)
  {
    if(to < from) /* (CHANGE2) */
    {
      double tmp;
      tmp = to; to = from; from = tmp;
      fprintf(stderr, "WARNING: exchanging from, to in llstat_new\n");
    }

    stat->from = from;
    stat->to = to;
    if( n_bask <= 0)  /* (CHANGE2) */
    {
      n_bask = 10;
      fprintf(stderr, "WARNING: setting n_bask=10 in llstat_new()\n");
    }
    stat->delta = (to-from)/(double) n_bask;
    stat->n_bask = n_bask;
    stat->low = 0;
    stat->high = 0;
    stat->num_jackknife = num_jackknife;
    stat->act_jackknife = 0;
    TFNE_MALLOC_AH(stat->sum_jackknife, _LLSTAT_NOM_, double *, "jackknife");
    for(moment=0; moment < _LLSTAT_NOM_; moment++)
    {
      TFNE_MALLOC_AH(stat->sum_jackknife[moment], num_jackknife, double, "jn");
      for(t=0; t<num_jackknife; t++)
	  stat->sum_jackknife[moment][t] = 0.0;
    }
    stat->magic = _LLSTAT_MAGIC_;

    for(t=0; t< _LLSTAT_NOM_ ; t++)
      stat->sum[t] = 0.0;
    TFN_MALLOC_AH(stat->table, n_bask, double, "Statistic");
    if(stat->table == NULL)
    {
      free(stat);
      stat = NULL;
    }
    else
      for(t=0; t<n_bask; t++)
	stat->table[t] = 0;
  }
  return(stat);
}

/** Cleares a statistic 'stat'     **/
void llstat_clear(llstat_t *stat)
{
  int t, moment;

#ifdef _LLSTAT_SEC_
  SEC_TEST(stat, _LLSTAT_MAGIC_, "Statistic");
#endif
  stat->low = 0;
  stat->high = 0;
  for(t=0; t < _LLSTAT_NOM_; t++)
    stat->sum[t] = 0.0;
  for(t=0; t < stat->n_bask; t++)
    stat->table[t] = 0;
  for(moment=0; moment < _LLSTAT_NOM_; moment++)
    for(t=0; t < stat->num_jackknife; t++)
      stat->sum_jackknife[moment][t] = 0.0;
  stat->act_jackknife = 0;
}

/** Deletes a statistic 'stat'     **/
void llstat_delete(llstat_t *stat)
{
  int moment;

#ifdef _LLSTAT_SEC_
  SEC_TEST(stat, _LLSTAT_MAGIC_, "Statistic");
#endif
  stat->magic = 0;
  free(stat->table);
  if(stat->num_jackknife > 0)
  {
    for(moment=0; moment < _LLSTAT_NOM_; moment++)
      free(stat->sum_jackknife[moment]);
    free(stat->sum_jackknife);
  }
  free(stat);
}

/** inserts a 'number' into a 'stat'. **/
void llstat_insert(llstat_t *stat, double number)
{
  int t;
  double value;
#ifdef _LLSTAT_SEC_
  SEC_TEST(stat, _LLSTAT_MAGIC_, "Statistic");
#endif
  value = 1.0;
  for(t=0; t< _LLSTAT_NOM_; t++)
  {
    stat->sum[t]+= value;                             /* raw statistics */ 
    if(stat->num_jackknife > 0)                 /* treat jackknife sums */
      stat->sum_jackknife[t][stat->act_jackknife] += value;
    value *= number;
  }
  if(number < stat->from)
    stat->low++;
  else if(number > stat->to)
    stat->high++;
  else if(number == stat->to)
    stat->table[stat->n_bask-1]++;
  else
    stat->table[(int) floor( (number - stat->from) / stat->delta)]++;
  if(stat->num_jackknife > 0)
  {
    stat->act_jackknife++;
    if(stat->act_jackknife >= stat->num_jackknife)
      stat->act_jackknife = 0;
  }
}

/******************* llstat_dens() *****************/
/** return density at 'x'                         **/
/** PARAMETERS: (*)= return-paramter              **/
/**          stat: histogram                      **/
/**             x: where                          **/
/** RETURNS:                                      **/
/**           density at 'x'                      **/
/******************************************/
double llstat_dens(llstat_t *stat, double x)
{
  int t;
  if(x < stat->from)
    return(0.0);
  else if(x > stat->to)
      return(0.0);
  else if(x == stat->to)
    t = stat->n_bask-1;
  else
    t= (int) floor( (x - stat->from) / stat->delta);
  return(stat->table[t])/(stat->sum[0]*stat->delta);
}

/** inserts a 'number' several 'times' into a 'stat'. **/
void llstat_insert_mult(llstat_t *stat, double number, double times)
{
  int t;
  double value;
#ifdef _LLSTAT_SEC_
  SEC_TEST(stat, _LLSTAT_MAGIC_, "Statistic");
#endif
  value = 1.0;
  for(t=0; t< _LLSTAT_NOM_; t++)
  {
    stat->sum[t]+= times* value;;                         /* raw statistics */ 
    value *= number;
  }
  if(number < stat->from)
    stat->low += times;
  else if(number > stat->to)
    stat->high += times;
  else if(number == stat->to)
    stat->table[stat->n_bask-1] +=times;
  else
    stat->table[(int) floor( (number - stat->from) / stat->delta)] +=times;
  if(stat->num_jackknife > 0)
  {
    fprintf(stderr, "llstat_insert_mult() does not support jackknife!\n");
    exit(1);
  }
}


/** RETURNS: No. of Elements in 'stat' **/
double llstat_no(llstat_t *stat)
{
#ifdef _LLSTAT_SEC_
  SEC_TEST(stat, _LLSTAT_MAGIC_, "Statistic");
#endif
  return(stat->sum[0]);
}


/** RETURNS: No. of Elements in 'stat' **/
double llstat_num(llstat_t *stat)
{
#ifdef _LLSTAT_SEC_
  SEC_TEST(stat, _LLSTAT_MAGIC_, "Statistic");
#endif
  return(stat->sum[0]);
}

/** RETURNS: Mean of Elements in 'stat' (0.0 if stat=empty) **/
double llstat_mean(llstat_t *stat)
{
#ifdef _LLSTAT_SEC_
  SEC_TEST(stat, _LLSTAT_MAGIC_, "Statistic");
#endif
  if(stat->sum[0] == 0)
    return(0.0);
  else
    return(stat->sum[1] / stat->sum[0]);
}

/** RETURNS: Mean of Elements in 'stat' (0.0 if stat=empty) **/
/**          without the entries in the jackknife box       **/
/**          'id jackknife'                                 **/
double llstat_mean_jackknife(llstat_t *stat, int id_jackknife)
{
  double num_elements;
#ifdef _LLSTAT_SEC_
  SEC_TEST(stat, _LLSTAT_MAGIC_, "Statistic");
#endif
  if(stat->num_jackknife == 0)
  {
    fprintf(stderr, "no jackknife defined!\n");
    exit(1);
  }
  if(stat->act_jackknife > id_jackknife)
    num_elements = floor(stat->sum[0]/stat->num_jackknife) + 1;
  else
    num_elements = floor(stat->sum[0]/stat->num_jackknife);
  if( (stat->sum[0]-num_elements) == 0)
    return(0.0);
  else
    return( (stat->sum[1] - stat->sum_jackknife[1][id_jackknife]) / 
	    (stat->sum[0]-num_elements));
}

/** RETURNS: Moment of Elements in 'stat' (0.0 if stat=empty) **/
/**          without the entries in the jackknife box         **/
/**          'id jackknife'                                   **/
double llstat_moment_jackknife(llstat_t *stat, int moment, int id_jackknife)
{
  double num_elements;
#ifdef _LLSTAT_SEC_
  SEC_TEST(stat, _LLSTAT_MAGIC_, "Statistic");
#endif
  if(stat->num_jackknife == 0)
  {
    fprintf(stderr, "no jackknife defined!\n");
    exit(1);
  }
  if(stat->act_jackknife > id_jackknife)
    num_elements = floor(stat->sum[0]/stat->num_jackknife) + 1;
  else
    num_elements = floor(stat->sum[0]/stat->num_jackknife);
  if( (stat->sum[0]-num_elements) == 0)
    return(0.0);
  else
    return( (stat->sum[moment] - stat->sum_jackknife[moment][id_jackknife]) / 
	    (stat->sum[0]-num_elements));
}

/******************* llcorr_median() **************/
/** Calculates median (accuray depend on the     **/
/** number of bins!)                             **/
/** PARAMETERS: (*)= return-paramter             **/
/**    stat: ...                                 **/
/** RETURNS:                                     **/
/**       median                                 **/
/**************************************************/
double llstat_median(llstat_t *stat)
{
  double median;
  double counter=stat->low;
  int t=0;
  int half = (int) (stat->sum[0]-1)/2;

  if( (stat->high > half)||(stat->low > half))
    return(-1.0);
#ifdef _LLSTAT_SEC_
  SEC_TEST(stat, _LLSTAT_MAGIC_, "Statistic");
#endif
  while(counter < half)                             /* look for meadian */
    counter += stat->table[t++];

  t--;
  if(counter == half)                    /* meadian is between two bins */
  {
    int t2 = t+1;

    while(stat->table[t2] ==0)
      t2++;

    median = ((double)(t+t2)*0.5+0.5)*stat->delta + stat->from;
  }
  else                                             /* median is in bin t */
    median = 
	(t + (double)(half-counter+stat->table[t])/stat->table[t])*stat->delta
        + stat->from;

  return(median);
}



/** RETURNS: Mean of moment n in 'stat' (0.0 if stat=empty) **/
double llstat_moment(llstat_t *stat, int n)
{
#ifdef _LLSTAT_SEC_
  SEC_TEST(stat, _LLSTAT_MAGIC_, "Statistic");
#endif
  if(stat->sum[0] == 0)
    return(0.0);
  else
    return(stat->sum[n] / stat->sum[0]);
}


/** Returns: Error of moment n in 'stat'  (0.0 if stat=empty) **/
double llstat_error_moment(llstat_t *stat, int n)
{
  double variance;
#ifdef _LLSTAT_SEC_
  SEC_TEST(stat, _LLSTAT_MAGIC_, "Statistic");
#endif
  if(stat->sum[0] == 0)
    return(0.0);
  else
      variance = (stat->sum[2*n] - stat->sum[n]*stat->sum[n]/stat->sum[0]) / 
	   (stat->sum[0]);
  return(sqrt(variance / stat->sum[0]) );
}


/** RETURNS: Statistical Variance of Elements in 'stat'  **/
/** (0.0 if stat has not a least two elements)           **/
double llstat_var(llstat_t *stat)
{
#ifdef _LLSTAT_SEC_
  SEC_TEST(stat, _LLSTAT_MAGIC_, "Statistic");
#endif
  if(stat->sum[0] == 0)
    return(0.0);
  else if(stat->sum[0] == 1)
    return(stat->sum[2]);
  else
    return((stat->sum[2] - stat->sum[1]*stat->sum[1]/stat->sum[0]) / 
	   (stat->sum[0]));
}


/******************* llstat_varvar() **************/
/** Statistical Variance of variance in 'stat'   **/
/** (0.0 if stat has not a least two elements)   **/
/** PARAMETERS: (*)= return-paramter             **/
/**    stat: ...                                 **/
/** RETURNS:                                     **/
/**       variance of variance                   **/
/**************************************************/
double llstat_varvar(llstat_t *stat)
{
  double x1, x2, x3, x4;
#ifdef _LLSTAT_SEC_
  SEC_TEST(stat, _LLSTAT_MAGIC_, "Statistic");
#endif
  if(stat->sum[0] < 2)
    return(0.0);

  x1 = stat->sum[1]/stat->sum[0];
  x2 = stat->sum[2]/stat->sum[0];
  x3 = stat->sum[3]/stat->sum[0];
  x4 = stat->sum[4]/stat->sum[0];

  return(x4 -4*x1*x1*x1*x1 + 8*x2*x1*x1 -4*x3*x1-x2*x2);
}

/** RETURNS: Standard error (65%) of Elements in 'stat'  **/
/** (0.0 if stat has not a least two elements)           **/
double llstat_error(llstat_t *stat)
{
  double variance;
#ifdef _LLSTAT_SEC_
  SEC_TEST(stat, _LLSTAT_MAGIC_, "Statistic");
#endif
  variance = llstat_var(stat);
  if((stat->sum[0]==0) || (variance <= 0.0))
    return(0.0);
  else
    return(sqrt(variance/(stat->sum[0]-1)));
}


/***************** llstat_errorvar() **************/
/** Statistical error of variance in 'stat'      **/
/** (0.0 if stat has not a least two elements)   **/
/** PARAMETERS: (*)= return-paramter             **/
/**    stat: ...                                 **/
/** RETURNS:                                     **/
/**       error of variance                      **/
/**************************************************/
double llstat_errorvar(llstat_t *stat)
{
  double variance;
#ifdef _LLSTAT_SEC_
  SEC_TEST(stat, _LLSTAT_MAGIC_, "Statistic");
#endif
  variance = llstat_varvar(stat);
  if((stat->sum[0]==0) || (variance <= 0.0))
    return(0.0);
  else
    return(sqrt(variance/stat->sum[0]));
}


/**************** llstat_write_log() **************/
/** Write statistic to file. The width of the    **/
/** interval of which is averaged is increased   **/
/** during output exponentially (i.e. on log     **/
/** scale the bins are equidistant)              **/                  
/** PARAMETERS: (*)= return-paramter             **/
/**    stat: ...                                 **/
/**    file: where to write                      **/
/**  factor: how fast interval increases         **/
/** RETURNS:                                     **/
/**       nothing                                **/
/**************************************************/
void llstat_write_log(llstat_t *stat, FILE *file, double factor)
{
  int  t, num_in_table;
  double bin0, bin1;
  double sum, t0, t1;

#ifdef _LLSTAT_SEC_
  SEC_TEST(stat, _LLSTAT_MAGIC_, "Statistic");
#endif
  if(((int) stat->sum[0]) == 0)
    fprintf(file, "# statistic is empty !\n");
  else
  {
    fprintf(file, "# n= %d , mean= %f +/-65%%= %f , var= %f\n", 
            llstat_no(stat),
            llstat_mean(stat), sqrt(llstat_var(stat)/llstat_no(stat)),
			llstat_var(stat));
    fprintf(file, "# %e elements to small, %e elements to big\n",
            stat->low, stat->high);
    num_in_table = ((int) stat->sum[0]);
    bin0=0.0; bin1=1.0;
    t0 = floor(bin0);
    t1 = floor(bin1);
    while(t0 < stat->n_bask)
    {
      sum = 0;
      for(t= (int) t0; (t<t1)&&(t1<stat->n_bask); t++)
	sum += stat->table[t];
      fprintf(file, "%e %e \n", sqrt( (stat->from+ stat->delta*(t0+0.5))*
			      (stat->from+ stat->delta*(t1-0.5))),
	     sum/(num_in_table*(t1-t0)*stat->delta));
      bin0=bin1;
      bin1 *= factor;
      if(bin1 < (bin0+1.0))
	  bin1 = bin0+1.0;
      t0 = floor(bin0);
      t1 = floor(bin1);
    }
  }
}


/** prints empiricalx s distribution 'stat' into a 'file'      **/
/* the file has to be open!                                 **/
/** the output is in gnuplot-fomrmat: the No. of datas, the **/
/** mean and the variance a printed as comments, the        **/
/** normalized density-function is printed as a table       **/
void llstat_write(llstat_t *stat, FILE *file)
{
  int t;
  double num_in_table;
  double value;

#ifdef _LLSTAT_SEC_
  SEC_TEST(stat, _LLSTAT_MAGIC_, "Statistic");
#endif
  if(( stat->sum[0]) == 0)
    fprintf(file, "# statistic is empty !\n");
  else
  {
    int tm;
    fprintf(file, "# moments:");
    for(tm=0; tm<_LLSTAT_NOM_; tm++)
    {
      fprintf(file, " %d: %f", tm, stat->sum[tm]/ stat->sum[0]);
      if(tm>0 && tm<=_LLSTAT_NOM_/2)
	fprintf(file, " %f", sqrt( (-stat->sum[tm]*stat->sum[tm]/stat->sum[0]+
			     stat->sum[2*tm])/(stat->sum[0]*stat->sum[0])));
    }
    fprintf(file, "\n");
    fprintf(file, "# n= %f , mean= %f +/-65%%= %f , var= %f\n", 
            llstat_no(stat),
            llstat_mean(stat), sqrt(llstat_var(stat)/llstat_no(stat)),
			llstat_var(stat));
    fprintf(file, "# %e elements to small, %e elements to big\n",
            stat->low, stat->high);
    num_in_table = ( stat->sum[0]) /*- stat->low - stat->high*/;/*CHANGE*/
    if(num_in_table == 0)
      fprintf(file, "# no numbers in table!\n");
    else
      for(t=0; t < stat->n_bask; t++)
      {
	value = ((double)stat->table[t])/(num_in_table);
        fprintf(file, "%e %e %e\n", 
		stat->from + (((double) t)+0.5) * stat->delta,
		value/stat->delta, 
		sqrt(value*(1-value)/num_in_table)/stat->delta);
      }
  }
}


/** Calculates x-value where statistic 'stat' has its maximum **/
/** and the value of the probability density there            **/
void llstat_max(llstat_t * stat, double *max_x, double *max_y)
{
  int t, max, num_in_table;
  double max_value;

  max = 0;
  max_value = 0;
  num_in_table = ((int) stat->sum[0]) /*- stat->low - stat->high*/;/*CHANGE*/
  for(t=0; t < stat->n_bask; t++)
	if( stat->table[t] > max)
	{
	  max = t;
	  max_value = stat->table[t];
	}
  *max_x = stat->from + (((double) max)+0.5) * stat->delta;
  *max_y = ((double)stat->table[max])/(((double)num_in_table)*stat->delta);
}


/** prints empirical distribution into standard-output      **/
/** the output is in gnuplot-fomrmat: the No. of datas, the **/
/** mean and the variance a printed as comments, the        **/
/** normalized density-function is printed as a table       **/
void llstat_print(llstat_t *stat)
{
  llstat_write(stat, stdout);
}

/** prints empirical distribution 'stat' into a 'file'      **/
/** the file has to be open!                                **/
/** The distribution is printed as Pair-Correlation-Fkt     **/
/** for a system with 'density' and 'number'= no.atoms*     **/
/** no.runs.                                                **/
/** the output is in gnuplot-fomrmat: the No. of datas, the **/
/** mean and the variance a printed as comments, the        **/
/** normalized density-function is printed as a table       **/
void llstat_w_pc(llstat_t *stat, FILE *file, double density, int number)
{
  int t, num_in_table;

#ifdef _LLSTAT_SEC_
  SEC_TEST(stat, _LLSTAT_MAGIC_, "Statistic");
#endif
  if(((int) stat->sum[0]) == 0)
    fprintf(file, "# statistic is empty !\n");
  else
  {
    fprintf(file, "# n= %d , mean= %f +/-65%%= %f , var= %f\n", 
			llstat_no(stat),
			llstat_mean(stat), sqrt(llstat_var(stat)/llstat_no(stat)),
			llstat_var(stat));
    fprintf(file, "# %e elements to small, %e elements to big\n",
	    stat->low, stat->high);
    num_in_table = ((int) stat->sum[0]) /*- stat->low - stat->high*/;/*CHANGE*/
    if(num_in_table == 0)
      fprintf(file, "# no numbers in table!\n");
    else
      for(t=0; t < stat->n_bask; t++)
      {
	double faktor = 4.0*PI*density*(pow(((double) t+1)*stat->delta, 3.0) -
			         pow(((double) t  )*stat->delta, 3.0))/3.0;
	fprintf(file, "%f %f\n", stat->from + (((double) t)+0.5) * stat->delta,
		((double) stat->table[t])/
		(((double)number)*faktor)   );
      }
  }
}

/** prints empirical distribution into standard-output      **/
/** as Pair-Correlation Funktion.                           **/
/** the output is in gnuplot-fomrmat: the No. of datas, the **/
/** mean and the variance a printed as comments, the        **/
/** normalized density-function is printed as a table       **/
void llstat_p_pc(llstat_t *stat, double density, int number)
{
  llstat_w_pc(stat, stdout, density, number);
}

/******************* llstat_int() *****************/
/** Integrates distribution                      **/
/** PARAMETERS: (*)= return-paramter             **/
/**        stat: distribution                    **/
/**    from, to: integration-range               **/
/** RETURNS:                                     **/
/**  integral value                              **/
/**************************************************/
double llstat_int(llstat_t *stat, double from, double to)
{
  double sum = 0.0;
  int t, t_from, t_to, num_in_table;

  /* calculate integrations-interval in basket-numbers */
  if(from < stat->from)
	from = stat->from;
  t_from = (int) ( (from - stat->from)/stat->delta + 0.9999);
  if(to > stat->to)
	to = stat->to;
  t_to = (int) ( (to - stat->from)/stat->delta); 

  /* integrate main part: */
  num_in_table = ((int) stat->sum[0]) /*- stat->low - stat->high*/;/*CHANGE*/
  for(t=t_from; t<t_to; t++)
	sum += stat->table[t];
  sum /= num_in_table;

  /* calculate part from overlapping boundaries: */
  if(t_from > 0)
	sum += stat->table[t_from-1]*
      ((stat->from + stat->delta*t_from) - from)/(stat->delta * num_in_table);
  if(t_to < stat->n_bask)
	sum += stat->table[t_to]*
	  (to - (stat->from + stat->delta*t_to))/(stat->delta * num_in_table);
  return(sum);
}


/** llstat_random_0_1 returns a 0-1 distributed number       **/
double llstat_random_0_1()
{
        return(drand48());
/*        return( ((double) rand())/(double) RAND_MAX);*/
/*      return( ((double) rand())/2147483648.0 );*/
/*      return( ((double)rand())/65535.0);*/
}

/** returns a N(0,1)-distributed random number per Box-Mueller- **/
/** Method, see B.J.T. Morgan, Elements of Simulation, S.78     **/
double llstat_random_norm()
{
  return(sqrt(-2*log(llstat_random_0_1()))*cos(2*PI*llstat_random_0_1()));
}

/** return integer random numberbetween 0 and n-1 **/
int llstat_random_0_n(int n)
{
/*      return( ((int) floor(zufallszahl_0_1()*n) ) );*/
/*        return(n-1- (rand() % n));*/
        return( (int) (n*drand48()));
}

/** llstat_init_rand() initialises the internal random **/
/** number generator using the actual time.            **/
void llstat_random_init()
{
        struct timeval zeit;  /* enthaelt die Zeit in sek, siehe sys/time.h */
        struct timezone zeitzone ;                      /* siehe sys/time.h */
        gettimeofday(&zeit,&zeitzone);                /* Hole aktuelle Zeit */
/*      printf("Zeit (sek):%d\n", (long)zeit.tv_sec% 214783648);*/      
        srand((long)(zeit.tv_sec % 214783648)); /* initialisiere Zufallsgen. */
}       


/*************** llcorr_new_jackknife() ***********/
/** creates correlation                          **/
/** PARAMETERS: (*)= return-paramter             **/
/**    from_x, to_x: x-range                     **/
/**    from_y, to_y: y-range                     **/
/**        n_bask_x: number of baskets in x      **/ 
/**        n_bask_y: number of baskets in y      **/ 
/**        num_jn_y: num. of jackknife entries   **/
/** RETURNS:                                     **/
/**  pointer to created correlation              **/
/**************************************************/
llcorr_t *llcorr_new_jackknife(double from_x, double to_x, 
			       double from_y, double to_y,
			       int n_bask_x, int n_bask_y, int num_jn_y)
{
  llcorr_t *corr;
  int t;

  TFE_MALLOC_AH(corr, llcorr_t, "correlation");	
  if(to_x < from_x)
  {
    double tmp;
    tmp = to_x; to_x = from_x; from_x = tmp;
    fprintf(stderr, "WARNING: exchanging from_x, to_x in llcorr_new\n");
  }
  if(to_y < from_y)
  {
    double tmp;
    tmp = to_y; to_y = from_y; from_y = tmp;
    fprintf(stderr, "WARNING: exchanging from_y, to_y in llcorr_new\n");
  }
  corr->from_x = from_x;
  corr->to_x = to_x;	
  if( n_bask_x <= 0)
  {
    n_bask_x = 10;
    fprintf(stderr, "WARNING: setting n_bask_y=10 in llstat_new()\n");
  }
  if( n_bask_y <= 0)
  {
    n_bask_y = 10;
    fprintf(stderr, "WARNING: setting n_bask_y=10 in llstat_new()\n");
  }
  corr->delta_x = (to_x-from_x)/(double) n_bask_x;
  corr->n_bask_x = n_bask_x;
  corr->low_x = llstat_new_jackknife(from_y, to_y, n_bask_y, num_jn_y);
  corr->high_x = llstat_new_jackknife(from_y, to_y, n_bask_y, num_jn_y);
  corr->num = 0;
  corr->magic = _LLCORR_MAGIC_;
  TFNE_MALLOC_AH(corr->stat, n_bask_x, llstat_t *, "corr: statistics");
  for(t=0; t<n_bask_x; t++)
	if( (corr->stat[t] = 
	     llstat_new_jackknife(from_y, to_y, n_bask_y, num_jn_y))== NULL)
	  FEHLER_AH("Out of memeory for creating correlations");
  return(corr);
}

/******************* llcorr_new() *****************/
/** creates correlation                          **/
/** PARAMETERS: (*)= return-paramter             **/
/**    from_x, to_x: x-range                     **/
/**    from_y, to_y: y-range                     **/
/**        n_bask_x: number of baskets in x      **/ 
/**        n_bask_y: number of baskets in y      **/ 
/** RETURNS:                                     **/
/**  pointer to created correlation              **/
/**************************************************/
llcorr_t *llcorr_new(double from_x, double to_x, double from_y, double to_y,
					 int n_bask_x, int n_bask_y)
{
  llcorr_t *corr;
  int t;

  TFE_MALLOC_AH(corr, llcorr_t, "correlation");	
  if(to_x < from_x)
  {
	double tmp;
	tmp = to_x; to_x = from_x; from_x = tmp;
	fprintf(stderr, "WARNING: exchanging from_x, to_x in llcorr_new\n");
  }
  if(to_y < from_y)
  {
	double tmp;
	tmp = to_y; to_y = from_y; from_y = tmp;
	fprintf(stderr, "WARNING: exchanging from_y, to_y in llcorr_new\n");
  }
  corr->from_x = from_x;
  corr->to_x = to_x;	
  if( n_bask_x <= 0)
  {
	n_bask_x = 10;
	fprintf(stderr, "WARNING: setting n_bask_y=10 in llstat_new()\n");
  }
  if( n_bask_y <= 0)
  {
	n_bask_y = 10;
	fprintf(stderr, "WARNING: setting n_bask_y=10 in llstat_new()\n");
  }
  corr->delta_x = (to_x-from_x)/(double) n_bask_x;
  corr->n_bask_x = n_bask_x;
  corr->low_x = llstat_new(from_y, to_y, n_bask_y);
  corr->high_x = llstat_new(from_y, to_y, n_bask_y);
  corr->num = 0;
  corr->magic = _LLCORR_MAGIC_;
  TFNE_MALLOC_AH(corr->stat, n_bask_x, llstat_t *, "corr: statistics");
  for(t=0; t<n_bask_x; t++)
	if( (corr->stat[t] = llstat_new(from_y, to_y, n_bask_y))== NULL)
	  FEHLER_AH("Out of memeory for creating correlations");
  return(corr);
}


/******************* llcorr_clear() ***************/
/** cleares correlation                          **/
/** PARAMETERS: (*)= return-paramter             **/
/**    corr: correlation                         **/
/** RETURNS:                                     **/
/**   nothing                                    **/
/**************************************************/
void llcorr_clear(llcorr_t *corr)
{
  int t;

#ifdef _LLSTAT_SEC_
  SEC_TEST(corr, _LLCORR_MAGIC_, "correlation");
#endif
  for(t=0; t<corr->n_bask_x; t++)
	llstat_clear(corr->stat[t]);
  llstat_clear(corr->low_x);
  llstat_clear(corr->high_x);
  corr->num = 0;
}


/******************* llcorr_delete() **************/
/** deletes correlation                          **/
/** PARAMETERS: (*)= return-paramter             **/
/**    corr: correlation                         **/
/** RETURNS:                                     **/
/**  nothing                                     **/
/**************************************************/
void llcorr_delete(llcorr_t *corr)
{
  int t;

#ifdef _LLSTAT_SEC_
  SEC_TEST(corr, _LLCORR_MAGIC_, "correlation");
#endif
  for(t=0; t<corr->n_bask_x; t++)
	llstat_delete(corr->stat[t]);
  free(corr->stat);
  llstat_delete(corr->low_x);
  llstat_delete(corr->high_x);
  free(corr);
}


/******************* llcorr_insert() **************/
/** inserts pair (x,y) into correlation          **/
/** PARAMETERS: (*)= return-paramter             **/
/**    corr: correlation                         **/
/**     x,y: values                              **/
/** RETURNS:                                     **/
/**   nothing                                    **/
/**************************************************/
void llcorr_insert(llcorr_t *corr, double x, double y)
{

#ifdef _LLSTAT_SEC_
  SEC_TEST(corr, _LLCORR_MAGIC_, "correlation");
#endif
  
  if(x < corr->from_x)
	llstat_insert(corr->low_x, y);
  else if (x > corr->to_x)
	llstat_insert(corr->high_x, y);
  else if ( x == corr->to_x)                            /* added, CHANGE3 */
	llstat_insert(corr->stat[corr->n_bask_x-1],	y);	
  else
	llstat_insert(corr->stat[(int) floor( (x-corr->from_x)/corr->delta_x)],
				  y);
  corr->num++;
}

/******************* llcorr_mean() **************/
/** Returns average value of correlation at 'x'  **/
/** PARAMETERS: (*)= return-paramter             **/
/**    corr: correlation                         **/
/**       x: where                              **/
/** RETURNS:                                     **/
/**   nothing                                    **/
/**************************************************/
double llcorr_mean(llcorr_t *corr, double x)
{
#ifdef _LLSTAT_SEC_
  SEC_TEST(corr, _LLCORR_MAGIC_, "correlation");
#endif
  
  if(x < corr->from_x)
    return(llstat_mean(corr->low_x));
  else if (x > corr->to_x)
    return(llstat_mean(corr->high_x));
  else if ( x == corr->to_x)                            /* added, CHANGE3 */
    return(llstat_mean(corr->stat[corr->n_bask_x-1]));	
  else
    return(llstat_mean(corr->stat[(int)floor((x-corr->from_x)/corr->delta_x)]));
}

/******************* llcorr_moment() **************/
/** Calculates n'th moment of correlation        **/
/** treated as probability distribution and      **/
/** its statistical error                        **/
/** PARAMETERS: (*)= return-paramter             **/
/**     corr: correlation                        **/
/**        n: order of moment                    **/
/** RETURNS:                                     **/
/**   n'th moment                                **/
/**************************************************/
double llcorr_moment(llcorr_t *corr, int n)
{
  double sum=0.0, x;
  int t;

  x = corr->from_x + corr->delta_x/2;
  for(t=0; t<corr->n_bask_x; t++)
  {
    sum += pow(x, (double) n)*llstat_mean(corr->stat[t]);
    x += corr->delta_x;
  }

  return(corr->delta_x*sum);
}

/******************* llcorr_write() ***************/
/** writes correlation to file, which must be    **/
/** open for writing                             **/
/** PARAMETERS: (*)= return-paramter             **/
/**    corr: correlation                         **/
/**    file: --                                  **/
/** RETURNS:                                     **/
/**  nothing                                     **/
/**************************************************/
void llcorr_write(llcorr_t *corr, FILE *file)
{
  int t;

#ifdef _LLSTAT_SEC_
  SEC_TEST(corr, _LLCORR_MAGIC_, "correlation");
#endif
  if(((int) corr->num) == 0)
    fprintf(file, "# statistic is empty !\n");
  else
  {
	fprintf(file, "# n= %d\n", corr->num);
    fprintf(file, "# %d elements to small = %f %f\n", llstat_no(corr->low_x),
			llstat_mean(corr->low_x), llstat_error(corr->low_x));
    fprintf(file, "# %d elements to big = %f %f\n", llstat_no(corr->high_x),
			llstat_mean(corr->high_x), llstat_error(corr->high_x));
	
    for(t=0; t<corr->n_bask_x; t++)
      if(llstat_no(corr->stat[t]) > 0)
	  fprintf(file, "%e %e %e  %d %d\n", 
		  corr->from_x + (((double) t)+0.5)*corr->delta_x,
		  llstat_mean(corr->stat[t]), llstat_error(corr->stat[t]),
		  t, (int) llstat_num(corr->stat[t]));
  }
}


/*************** llcorr_write_log() ***************/
/** writes correlation to file, which must be    **/
/** open for writing         The width of the    **/
/** interval of which is averaged is increased   **/
/** during output exponentially (i.e. on log     **/
/** scale the bins are equidistant)              **/                  
/** PARAMETERS: (*)= return-paramter             **/
/**    corr: correlation                         **/
/**    file: --                                  **/
/**  factor: who fast interval increases         **/
/** RETURNS:                                     **/
/**  nothing                                     **/
/**************************************************/
void llcorr_write_log(llcorr_t *corr, FILE *file, double factor)
{
  int t;
  double sum, bin0, bin1, sum_norm;
  double t0, t1;
  llstat_t *value;

  value = llstat_new(0,1,2);

#ifdef _LLSTAT_SEC_
  SEC_TEST(corr, _LLCORR_MAGIC_, "correlation");
#endif
  if(((int) corr->num) == 0)
    fprintf(file, "# statistic is empty !\n");
  else
  {
    fprintf(file, "# n= %d\n", corr->num);
    fprintf(file, "# %d elements to small = %f %f\n", llstat_no(corr->low_x),
			llstat_mean(corr->low_x), llstat_error(corr->low_x));
    fprintf(file, "# %d elements to big = %f %f\n", llstat_no(corr->high_x),
			llstat_mean(corr->high_x), llstat_error(corr->high_x));
	
    bin0=0.0; bin1=1.0;
    t0 = floor(bin0);
    t1 = floor(bin1);
    while(t0 < corr->n_bask_x)
    {
      llstat_clear(value);
      sum = 0;
      sum_norm = 0;
      for(t= (int) t0; (t<t1)&&(t1<corr->n_bask_x); t++)
      {
	if(llstat_num(corr->stat[t])>0)
	    llstat_insert_mult(value,llstat_mean(corr->stat[t]),
			       llstat_num(corr->stat[t]));
	sum += llstat_num(corr->stat[t])*llstat_mean(corr->stat[t]);
	sum_norm += llstat_num(corr->stat[t]);
      }
      if(llstat_num(value) > 0)
        fprintf(file, "%e %e %e %d %f\n",  sqrt((corr->from_x+ corr->delta_x*(t0+0.5))*
				       (corr->from_x+ corr->delta_x*(t1-0.5))),
		llstat_mean(value), llstat_error(value),
		llstat_num(value), sum_norm);
      bin0=bin1;
      bin1 *= factor;
      if(bin1 < (bin0+1.0))
	  bin1 = bin0+1.0;
      t0 = floor(bin0);
      t1 = floor(bin1);
    }		 
  }
  llstat_delete(value);
}


/******************* llcorr_print() ***************/
/** writes correlation to stdout                 **/
/** PARAMETERS: (*)= return-paramter             **/
/**    corr: correlation                         **/
/** RETURNS:                                     **/
/**  nothing                                     **/
/**************************************************/
void llcorr_print(llcorr_t *corr)
{
  llcorr_write(corr, stdout);
}


/***************** llcorr_add_stat() **************/
/** adds values of all colums of 'stat' to 'corr'**/
/** both must have same layout in x direction.   **/
/** PARAMETERS: (*)= return-paramter             **/
/**    corr: correlation                         **/
/**     x,y: values                              **/
/** RETURNS:                                     **/
/**   0 -> OK, 1 -> not same layout              **/
/**************************************************/
int llcorr_add_stat(llcorr_t *corr, llstat_t *stat)
{
  int t, num_in_table;

  /* have corr and stat the same layout in x direction ? */
  if( (stat->from != corr->from_x) || (stat->to != corr->to_x) ||
	  (stat->n_bask != corr->n_bask_x))
	return(1);

  num_in_table = ((int) stat->sum[0]) /*- stat->low - stat->high*/; /*CHANGE*/
  if(num_in_table > 0)
  {
    for(t=0; t<stat->n_bask; t++)                               /* add stat */
      llstat_insert(corr->stat[t], 
	   ((double)stat->table[t])/(((double)num_in_table)*stat->delta));
    corr->num ++;
  }
  return(0);
}

/******************** llstat_binder() *****************/
/** Calculates binder cumulant 0.5*(3-<m4>/<m2*m2>   **/
/** from statistics for m4 and m2.                   **/
/** PARAMETERS: (*)= return-paramter                 **/
/**   m2_stat, m4_stat: input statistics             **/
/**        (*) error65: standard error of cumulant   **/
/** RETURNS:                                         **/
/**   binder-cumulant                                **/
/******************************************************/
double llstat_binder(llstat_t *m2_stat, llstat_t *m4_stat, double *error65)
{
  double m2, m4, dm2, dm4;

  m2 = llstat_mean(m2_stat);
  m4 = llstat_mean(m4_stat);
  dm2 = llstat_error(m2_stat)*llstat_error(m2_stat);
  dm4 = llstat_error(m4_stat)*llstat_error(m4_stat);

  /* error with usual 'Fehlerfortpflanzung' */
  *error65 = sqrt(fabs(1/(4*m2*m2*m2*m2))*dm4 + 
		  fabs(m4*m4/(m2*m2*m2*m2*m2*m2))*dm2);
  return(0.5*(3.0-m4/(m2*m2)));
}

/******************** llstat_binder2() ****************/
/** Calculates binder cumulant 0.5*(3-<m4>/<m2*m2>)  **/
/** from statistics for m.                           **/
/** PARAMETERS: (*)= return-paramter                 **/
/**        m_stat: input statistic                   **/
/**   (*) error65: standard error of cumulant        **/
/** RETURNS:                                         **/
/**   binder-cumulant                                **/
/******************************************************/
double llstat_binder2(llstat_t *m_stat, double *error65)
{
  double m2, m4, sigma2, sigma4, value;
  int id_jack;
  llstat_t *binder;

  if(m_stat->num_jackknife > 0)      /* use jackknife analysis ? */
  {
    binder = llstat_new(0, 1, 2);
    for(id_jack =0; id_jack < m_stat->num_jackknife; id_jack++)
    {
      m2 = llstat_moment_jackknife(m_stat, 2, id_jack);
      m4 = llstat_moment_jackknife(m_stat, 4, id_jack);
      llstat_insert(binder, 0.5*(3.0-m4/(m2*m2)));
    }
    value = llstat_mean(binder);
    *error65 = sqrt( (m_stat->num_jackknife-1)*llstat_var(binder));
    llstat_delete(binder);
    return(value);
  }

  m2 = m_stat->sum[2]/m_stat->sum[0];
  m4 = m_stat->sum[4]/m_stat->sum[0];

  sigma2 =  (m_stat->sum[4]- m_stat->sum[2]*m_stat->sum[2]/m_stat->sum[0])/
               (m_stat->sum[0]*(m_stat->sum[0]-1.0));
  sigma4 =   (m_stat->sum[8]- m_stat->sum[4]*m_stat->sum[4]/m_stat->sum[0])/
               (m_stat->sum[0]*(m_stat->sum[0]-1.0));

  /* error with usual 'error propagation' */
  *error65 = sqrt(fabs(1/(4*m2*m2*m2*m2))*sigma4 + 
		  fabs(m4*m4/(m2*m2*m2*m2*m2*m2))*sigma2);
  return(0.5*(3.0-m4/(m2*m2)));
}
